import collections
import json
from typing import no_type_check

import piemap.projections as pr
from piemap import PIPE, SEMI, log

NEEDED_NUMBER_OF_AXIS_MAX = 16
NULL_STR_REP = 'NULL'
REC_SEP = ';'
ROW_SEP = '\n'


@no_type_check
def dumps(config):
    """Dump configuration into string."""
    return json.dumps(config)


@no_type_check
def dump(config, handle):
    """Dump configuration into JSON file via handle."""
    return json.dump(config, handle)


@no_type_check
def loads(text):
    """Load configuration from string."""
    return NotImplemented


@no_type_check
def load(handle):
    """Load configuration from JSON file handle."""
    return json.load(handle)


@no_type_check
def default_linear():
    """DRY."""
    return {
        'AXIS_INDEX': '',
        'AXIS_NAME': 'Dimension',
        'AXIS_TYPE': 'LINEAR',
        'AXIS_MIN': 0.00,
        'AXIS_LIMIT': 0.80,
        'AXIS_MAX': 1.00,
        'AXIS_LIMIT_FOLDED': False,
        'AXIS_MIN_FOLDED': False,
        'AXIS_VALUE': NULL_STR_REP,
        'AXIS_UNIT': '1',
        'AXIS_META': '',
    }


@no_type_check
def default_keys():
    """DRY."""
    return list(default_linear().keys())


@no_type_check
def default_linear_values():
    """DRY."""
    return list(default_linear().values())


@no_type_check
def default_folded():
    """DRY."""
    return {
        'AXIS_INDEX': '',
        'AXIS_NAME': 'DimensionFolded',
        'AXIS_TYPE': 'FOLDED',
        'AXIS_MIN': 0.00,
        'AXIS_LIMIT': 0.80,
        'AXIS_MAX': 1.00,
        'AXIS_LIMIT_FOLDED': False,
        'AXIS_MIN_FOLDED': False,
        'AXIS_VALUE': NULL_STR_REP,
        'AXIS_UNIT': 'dB',
        'AXIS_META': '',
    }


@no_type_check
def default_folded_values():
    """DRY."""
    return list(default_folded().values())


@no_type_check
def unsafe(text):
    """TODO(sthagen) keep/drop list applies here."""
    return text


@no_type_check
def is_numeric(text):
    """Migration artifact."""
    try:
        _ = float(text)
        return True
    except (TypeError, ValueError):
        return False


@no_type_check
def maybe_int(value):
    """Rococo."""
    if float(value) == float(int(float(value))):
        return int(value)

    return value


@no_type_check
def compact_value(text):
    """HACK A DID ACK ..."""
    if not is_numeric(text):
        return text

    try:
        as_int = int(text)
        return as_int
    except (TypeError, ValueError):
        return float(text)


@no_type_check
def example_default():
    """Maybe not wanted but matches reference behavior."""
    some_axis_maps = [default_linear()]
    some_axis_maps[0]['AXIS_INDEX'] = 0

    some_axis_maps.append(default_linear())
    some_axis_maps[1]['AXIS_INDEX'] = 1
    some_axis_maps[1]['AXIS_NAME'] = 'Dimension2'

    some_axis_maps.append(default_folded())
    some_axis_maps[2]['AXIS_INDEX'] = 2

    return some_axis_maps, ['Default used, since no input given.']


@no_type_check
def parse(text: str):
    """Parse the DSL contained in text."""
    if not text.strip():
        return example_default()

    has_index_collision = False
    has_index_order_mismatch = False

    info_queue, some_axis_maps = [], []
    axis_values_rows_req_string: str = unsafe(text)
    axis_values_rows_req = [row.strip() for row in axis_values_rows_req_string.split(ROW_SEP) if row.strip()]
    n_axis_rows_req = len(axis_values_rows_req)
    for n, row_string in enumerate(
        axis_values_rows_req[:NEEDED_NUMBER_OF_AXIS_MAX]
    ):  # Was: array_slice(axis_values_rows_req, 0, NEEDED_NUMBER_OF_AXIS_MAX)
        axis_values = default_folded_values() if ';FOLDED;' in row_string else default_linear_values()
        axis_values_cand = row_string.split(REC_SEP)
        log.info(f'__{"FOLDED" if ";FOLDED;" in row_string else "NO_FOLD"}__')
        log.info(f'has axis_values[{len(axis_values)}]({axis_values})')
        if len(axis_values) >= len(axis_values_cand):
            kks = default_keys()
            for i, v in enumerate(axis_values_cand):
                if v != '':
                    be_v = compact_value(v)
                    log.debug(f'[{i}]( == {kks[i]}): ({axis_values[i]}) --> ({be_v})')
                    axis_values[i] = be_v
                else:
                    log.debug(f'[{i}]( == {kks[i]}): ({axis_values[i]}) kept ({axis_values[i]})')
        if len(axis_values) > len(axis_values_cand):
            kks = default_keys()
            for i, v in enumerate(axis_values[len(axis_values_cand) :], start=len(axis_values_cand)):
                log.debug(f'[{i}]( == {kks[i]}): ({axis_values[i]}) untouched ({axis_values[i]})')

        log.info(f'has axis candidate[{len(axis_values_cand)}]({axis_values_cand})')
        axis_map = dict(zip(default_keys(), axis_values))
        log.info(f'... axis map is ({axis_map})')
        numeric_axis_types = ('LINEAR', 'FOLDED')
        if axis_map['AXIS_INDEX'] == '':
            log.info(f"has index idea({axis_map['AXIS_INDEX']}) is False?")
            log.info(f'... axis map is ({axis_map})')
            axis_map['AXIS_INDEX'] = n
            i_cfc = axis_map['AXIS_INDEX']
        else:
            print()
            log.info(f"has index idea({axis_map['AXIS_INDEX']}) is True")
            index_cand = str(axis_map['AXIS_INDEX'])
            if is_numeric(index_cand) and float(index_cand) == float(int(float(index_cand))):
                log.info(f'is_numeric({index_cand}) is True')
                i_cfc = str(maybe_int(index_cand))
                axis_map['AXIS_INDEX'] = int(i_cfc)
                if index_cand != i_cfc:  # Was !== in PHP
                    info_queue.append(f'NOK index ({index_cand}) requested, accepted as ({i_cfc})')
                else:
                    info_queue.append(f' OK index ({index_cand}) requested, accepted as ({i_cfc})')
            else:
                log.info(f'is_numeric({index_cand}) is False')
                i_cfc = str(n)
                info_queue.append(
                    f'NOK invalid index ({index_cand}) requested, accepted per parsing order as ({i_cfc})'
                )
        if axis_map['AXIS_TYPE'] in numeric_axis_types:
            if axis_map['AXIS_TYPE'] == 'LINEAR':
                if is_numeric(axis_map['AXIS_LIMIT']) and is_numeric(axis_map['AXIS_MAX']):
                    axis_map['AXIS_MIN'] = maybe_int(
                        pr.min_from_limit_max(axis_map['AXIS_LIMIT'], axis_map['AXIS_MAX'])
                    )
                else:
                    info_queue.append(
                        f"NOK limit({axis_map['AXIS_LIMIT']}) and max({axis_map['AXIS_MAX']}) not both numeric,"
                        f" ignored {axis_map['AXIS_TYPE'].lower()} axis at index ({i_cfc})"
                    )
                if axis_map['AXIS_VALUE'] != NULL_STR_REP and not is_numeric(axis_map['AXIS_VALUE']):
                    axis_map['AXIS_VALUE'] = NULL_STR_REP
            else:  # axis_map['AXIS_TYPE'] == 'FOLDED':
                log.info(f'... claim of FOLDED is True? Axis map is ({axis_map})')
                if is_numeric(axis_map['AXIS_LIMIT']) and is_numeric(axis_map['AXIS_MAX']):
                    axis_map['AXIS_MIN'] = maybe_int(
                        pr.min_from_limit_max(axis_map['AXIS_LIMIT'], axis_map['AXIS_MAX'])
                    )
                    axis_map['AXIS_LIMIT_FOLDED'] = maybe_int(
                        pr.limit_folded_from_limit_max(axis_map['AXIS_LIMIT'], axis_map['AXIS_MAX'])
                    )
                    axis_map['AXIS_MIN_FOLDED'] = maybe_int(
                        pr.min_folded_from_limit_max(axis_map['AXIS_LIMIT'], axis_map['AXIS_MAX'])
                    )
                else:
                    info_queue.append(
                        f"NOK limit({axis_map['AXIS_LIMIT']}) and max({axis_map['AXIS_MAX']}) not both numeric,"
                        f' ignored folded axis at index ({i_cfc})'
                    )

        some_axis_maps.append(axis_map)

    n_axis_rows = len(some_axis_maps)
    if n_axis_rows_req > n_axis_rows:
        info_queue.append(
            f'{n_axis_rows_req} dimensions requested, but only {n_axis_rows} accepted.'
            f' Maximum is {NEEDED_NUMBER_OF_AXIS_MAX}'
        )

    best_effort_re_order_map = {}
    collect_index_cand_list = []
    for x, data in enumerate(some_axis_maps):
        index_cand = data['AXIS_INDEX']
        if is_numeric(index_cand):
            i_cfc = int(index_cand)
        else:
            i_cfc = x
        collect_index_cand_list.append(i_cfc)
        if not is_numeric(index_cand) or i_cfc != index_cand or index_cand < 0 or index_cand >= n_axis_rows:
            has_index_collision = True
            conflict_reason = 'NO_INTEGER'
            if i_cfc != index_cand:
                conflict_reason = 'DC_INTEGER'

            if is_numeric(index_cand) and index_cand < 0:
                conflict_reason = 'LT_ZERO'

            elif is_numeric(index_cand) and index_cand >= n_axis_rows:
                conflict_reason = 'GT_NROW'

            info_queue.append(
                f'Conflicting index rules. Failing candidate is ({index_cand}), reason is {conflict_reason}'
            )
            some_axis_maps[x]['AXIS_INDEX'] = x

        if index_cand != x:
            has_index_order_mismatch = True
            info_queue.append(f'Index positions not ordered. Misplaced candidate is {index_cand}, found at {x}')

        best_effort_re_order_map[index_cand] = data

    collect_index_cand_set = set(collect_index_cand_list)
    if len(collect_index_cand_set) != len(collect_index_cand_list):
        has_index_collision = True
        hist = collections.Counter(collect_index_cand_list)
        blame_list = []
        for xx, nn in hist.items():
            if nn != 1:
                blame_list.append(xx)

        blame_list.sort()
        info_queue.append(
            f'Conflicting index positions. Failing candidate/s is/are [{", ".join(str(x) for x in blame_list)}],'
            ' reason is NONUNIQUE_INDEX'
        )

    if not has_index_collision and has_index_order_mismatch:
        some_axis_maps = []
        for k, data in sorted(best_effort_re_order_map.items()):
            some_axis_maps.append(data)

    """
    $normalizedInputDataRows = array();
    foreach($someAxisMaps as $x => $data) {
        $axisValues = array_values($data);
        $normalizedInputDataRows[] = implode(";",$axisValues);
    }
    //DEBUG echo '<pre>ReAssembledRows:'."\n".print_r($normalizedInputDataRows,True).'</pre>';
    $normalizedInputDataString = implode("\n",$normalizedInputDataRows);
    //DEBUG echo '<pre>ReAssembledNormalizedInput:'."\n".print_r($normalizedInputDataString,True).'</pre>';
    echo 'AxisSpecTest: '."\n";
    echo '<form style="display:inline;" method="post" action="'.$_SERVER['PHP_SELF'].'">'."\n";
    echo '<textarea style="font-sice:small;" cols="80" rows="16" name="AXIS_SPEC_ROWS">'.$normalizedInputDataString. \
      </textarea>'."\n";
    echo '<input type="submit" name="Subme" value="parse" />'."\n";
    echo '</form>'.'<br />'."\n";
    echo '[<a href="'.$_SERVER['PHP_SELF'].'">RESET</a>] to some default to get started.<br />'."\n";
    //echo 'Implicit Keys: '.implode(';',$axisDefaultKeys).'<br />'."\n";
    echo '<pre>';
    echo 'Testing Module: '.$_SERVER['PHP_SELF']."\n";
    $infoQueue[] = 'TestInput: AXIS_SPEC_ROWS=<pre>'."\n".$axisValuesRowsReqString.'</pre>'."\n";
    echo '  TestOutput[0]:'."\n";
    echo '$someAxisMaps='."\n";
    //DEBUG echo print_r($someAxisMaps,True);
    echo '</pre>';
    echo '<table style="width:75%;"><tr><th>Laufd.Nr.</th><th>Name</th><th>Type</th><th>Min</th><th>Limit</th>\
      <th>Max</th><th>LimitFolded</th><th>MinFolded</th><th>Value</th><th>Unit</th></tr>'."\n";
    foreach($someAxisMaps as $i => $data) {
        $displayRow = array_values($data);
        echo '<tr><td>';
        echo implode('</td><td>',$displayRow);
        echo '</td></tr>';
    }
    echo '</table>'."\n";
    if($infoQueue) {
        echo '<h2>Info:</h2>';
        echo '<ul><li>';
        echo implode('</li><li>',$infoQueue);
        echo '</li></ul>';
    }
    echo '</body>'."\n";
    echo '</html>'."\n";
    """
    return some_axis_maps, info_queue


@no_type_check
def table(axis_maps, info_queue) -> str:
    """Display axis maps as markdown table."""
    normalized_input_data_rows = [SEMI.join(str(v) for v in axis.values()) for axis in axis_maps]
    log.debug('ReAssembledRows:')
    for row in normalized_input_data_rows:
        log.debug(row)
    normalized_input_data_string = '\n'.join(normalized_input_data_rows)
    log.debug('ReAssembledNormalizedInput:')
    log.debug(normalized_input_data_string)
    table_headers = ('ListNo.', 'Name', 'Type', 'Min', 'Limit', 'Max', 'LimitFolded', 'MinFolded', 'Value', 'Unit')
    table_header_str_row = PIPE.join(table_headers)
    table_body_str_rows = []
    for i, data in enumerate(axis_maps, start=1):
        display_row = PIPE.join(str(d) for d in data.values())
        table_body_str_rows.append(display_row)
    if info_queue:
        log.warning('InfoQueue:')
        for diagnostic in info_queue:
            log.warning(f'- {diagnostic}')
    return '\n'.join([table_header_str_row] + table_body_str_rows)



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
import hypothesis.strategies as st
from hypothesis import given

import piemap.dsl as dsl


def test_dumps_stub_ok():
    assert dsl.dumps({}) == '{}'


def test_loads_stub_ok():
    assert dsl.loads('') is NotImplemented


def test_is_numeric_true_ok():
    assert dsl.is_numeric('42') is True


def test_is_numeric_false_ok():
    assert dsl.is_numeric('fourtytwo') is False


def test_parse_empty():
    default_text_0 = ''

    default_generated_0 = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'Dimension',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 1,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'Dimension2',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 2,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'DimensionFolded',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 'NULL',
        },
    ]

    default_diagnostics_0 = ['Default used, since no input given.']

    assert dsl.parse(default_text_0) == (default_generated_0, default_diagnostics_0)


def test_parse_dim_one_linear():
    default_text_1 = """\
    0;D1F;FOLDED;;12;15;6;dB
    """

    default_parsed_1 = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D1F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 'NULL',
        },
    ]

    default_diagnostics_1 = [' OK index (0) requested, accepted as (0)']

    assert dsl.parse(default_text_1) == (default_parsed_1, default_diagnostics_1)


def test_parse_dim_two_linear():
    default_text_2 = """\
    ;L 0;;;1;2;;;NULL;
    ;L 1;;;1;2;;;-0.1;
    """

    default_parsed_2 = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'L 0',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 1,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'L 1',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': -0.1,
        },
    ]

    default_diagnostics_2 = []

    assert dsl.parse(default_text_2) == (default_parsed_2, default_diagnostics_2)


def test_parse_dim_three_folded():
    default_text_3 = """\
    ;F 1;FOLDED;;1;2;;;2;
    ;F 2;FOLDED;;1;2;;;0;
    ;F 3;FOLDED;;1;2;;;0.25;
    """

    default_parsed_3 = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': 3,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': 4.666666666666667,
            'AXIS_NAME': 'F 1',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 2,
        },
        {
            'AXIS_INDEX': 1,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': 3,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': 4.666666666666667,
            'AXIS_NAME': 'F 2',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 0,
        },
        {
            'AXIS_INDEX': 2,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': 3,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': 4.666666666666667,
            'AXIS_NAME': 'F 3',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 0.25,
        },
    ]

    default_diagnostics_3 = []

    assert dsl.parse(default_text_3) == (default_parsed_3, default_diagnostics_3)


def test_parse_dim_two_linear_unordered():
    default_text_2_unordered = """\
    1;L 0;;;1;2;;;NULL;
    0;L 1;;;1;2;;;-0.1;
    """

    default_parsed_2_unordered = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'L 1',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': -0.1,
        },
        {
            'AXIS_INDEX': 1,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'L 0',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
    ]

    default_diagnostics_2_unordered = [
        ' OK index (1) requested, accepted as (1)',
        ' OK index (0) requested, accepted as (0)',
        'Index positions not ordered. Misplaced candidate is 1, found at 0',
        'Index positions not ordered. Misplaced candidate is 0, found at 1',
    ]

    assert dsl.parse(default_text_2_unordered) == (default_parsed_2_unordered, default_diagnostics_2_unordered)


def test_parse_dim_two_linear_collision():
    default_text_2_collision = """\
    333;L 0C;;;1;2;;;NULL;
    333;L 1C;;;1;2;;;-0.1;;
    """

    default_parsed_2_collision = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'L 0C',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 1,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'L 1C',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': -0.1,
        },
    ]

    default_diagnostics_2_collision = [
        ' OK index (333) requested, accepted as (333)',
        ' OK index (333) requested, accepted as (333)',
        'Conflicting index rules. Failing candidate is (333), reason is GT_NROW',
        'Index positions not ordered. Misplaced candidate is 333, found at 0',
        'Conflicting index rules. Failing candidate is (333), reason is GT_NROW',
        'Index positions not ordered. Misplaced candidate is 333, found at 1',
        'Conflicting index positions. Failing candidate/s is/are [333], reason is NONUNIQUE_INDEX',
    ]

    assert dsl.parse(default_text_2_collision) == (default_parsed_2_collision, default_diagnostics_2_collision)


def test_parse_d1f_ok():
    text = """\
    0;D1FX;FOLDED;;12;15;;;6;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D1FX',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 6,
        },
    ]

    diag = [' OK index (0) requested, accepted as (0)']

    assert dsl.parse(text) == (ast, diag)


def test_parse_d2f_ok():
    text = """\
    1;D2F;FOLDED;;12;15;;;12;V
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D2F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'V',
            'AXIS_VALUE': 12,
        },
    ]

    diag = [
        ' OK index (1) requested, accepted as (1)',
        'Conflicting index rules. Failing candidate is (1), reason is GT_NROW',
        'Index positions not ordered. Misplaced candidate is 1, found at 0',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_d3f_ok():
    text = """\
    2;D3F;FOLDED;;12;15;NULL;1
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D3F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        ' OK index (2) requested, accepted as (2)',
        'Conflicting index rules. Failing candidate is (2), reason is GT_NROW',
        'Index positions not ordered. Misplaced candidate is 2, found at 0',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_d4f_ok():
    text = """\
    ;D4F;FOLDED;;12;15;16;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D4F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d5f_ok():
    text = """\
    ;D5F;FOLDED;;12;15;18;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D5F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d6f_ok():
    text = """\
    ;D6F;FOLDED;;12;15;22;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D6F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d7b_ok():
    text = """\
    ;D7B;BIMONOTONE;0;80;100;;;40;%;SHOW_MIN
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 80,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 100,
            'AXIS_META': 'SHOW_MIN',
            'AXIS_MIN': 0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D7B',
            'AXIS_TYPE': 'BIMONOTONE',
            'AXIS_UNIT': '%',
            'AXIS_VALUE': 40,
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d7b_wrong_index_type_ok():
    wrong_index_type_text = """\
    WRONG;D7B;BIMONOTONE;0;80;100;;;40;%;SHOW_MIN
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 80,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 100,
            'AXIS_META': 'SHOW_MIN',
            'AXIS_MIN': 0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D7B',
            'AXIS_TYPE': 'BIMONOTONE',
            'AXIS_UNIT': '%',
            'AXIS_VALUE': 40,
        },
    ]

    wrong_index_type_diag = [
        'NOK invalid index (WRONG) requested, accepted per parsing order as (0)',
        'Conflicting index rules. Failing candidate is (WRONG), reason is DC_INTEGER',
        'Index positions not ordered. Misplaced candidate is WRONG, found at 0',
    ]

    assert dsl.parse(wrong_index_type_text) == (ast, wrong_index_type_diag)


def test_parse_d8l_ok():
    text = """\
    ;D8L;;;1;2;;;0.5;#
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D8L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '#',
            'AXIS_VALUE': 0.5,
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d9f_ok():
    text = """\
    ;D9F;FOLDED;;12;15;;;12;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': 18,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 7,
            'AXIS_MIN_FOLDED': 23,
            'AXIS_NAME': 'D9F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 12,
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d10l_ok():
    text = """\
    ;D10L;;;1;2;;;0.25;ms
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 1,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 2,
            'AXIS_META': '',
            'AXIS_MIN': -0.6666666666666666,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D10L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 'ms',
            'AXIS_VALUE': 0.25,
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d11l_ok():
    text = """\
    ;D11L;;;2;1;;;0.8;kbit/s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 2,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 3.6666666666666665,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D11L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 'kbit/s',
            'AXIS_VALUE': 0.8,
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_d12l_ok():
    text = """\
    ;D12L;;;0.8;1;;;NULL;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_value_neither_null_nor_numeric_ok():
    text = """\
    ;D12L;;;0.8;1;;;wrong_value;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = []

    assert dsl.parse(text) == (ast, diag)


def test_parse_limit_not_numeric_ok():
    text = """\
    ;D12L;;;non_numeric_limit;1;;;NULL;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 'non_numeric_limit',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        'NOK limit(non_numeric_limit) and max(1) not both numeric, ignored linear axis at index (0)',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_folded_limit_not_numeric_ok():
    text = """\
    ;D9F;FOLDED;;non_numeric_limit;15;;;12;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 'non_numeric_limit',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 15,
            'AXIS_META': '',
            'AXIS_MIN': 0.0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D9F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 12,
        },
    ]

    diag = [
        'NOK limit(non_numeric_limit) and max(15) not both numeric, ignored folded axis at index (0)',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_max_not_numeric_ok():
    text = """\
    ;D12L;;;0.8;non_numeric_max;;;NULL;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 'non_numeric_max',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_META': '',
            'AXIS_MIN': 0.0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        'NOK limit(0.8) and max(non_numeric_max) not both numeric, ignored linear axis at index (0)',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_folded_max_not_numeric_ok():
    text = """\
    ;D9F;FOLDED;;12;non_numeric_max;;;12;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 12,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 'non_numeric_max',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_META': '',
            'AXIS_MIN': 0.0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D9F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 12,
        },
    ]

    diag = [
        'NOK limit(12) and max(non_numeric_max) not both numeric, ignored folded axis at index (0)',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_limit_and_max_not_numeric_ok():
    text = """\
    ;D12L;;;non_numeric_limit;non_numeric_max;;;NULL;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 'non_numeric_limit',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 'non_numeric_max',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_META': '',
            'AXIS_MIN': 0.0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        'NOK limit(non_numeric_limit) and max(non_numeric_max) not both numeric, ignored linear axis at index (0)',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_folded_limit_and_max_not_numeric_ok():
    text = """\
    ;D9F;FOLDED;;non_numeric_limit;non_numeric_max;;;12;dB
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 'non_numeric_limit',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 'non_numeric_max',  # This is maybe not what we want ... TODO(sthagen)
            'AXIS_META': '',
            'AXIS_MIN': 0.0,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D9F',
            'AXIS_TYPE': 'FOLDED',
            'AXIS_UNIT': 'dB',
            'AXIS_VALUE': 12,
        },
    ]

    diag = [
        'NOK limit(non_numeric_limit) and max(non_numeric_max) not both numeric, ignored folded axis at index (0)',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_index_as_float_ok():
    text = """\
    0.333;D12L;;;0.8;1;;;NULL;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        'NOK invalid index (0.333) requested, accepted per parsing order as (0)',
        'Conflicting index rules. Failing candidate is (0.333), reason is DC_INTEGER',
        'Index positions not ordered. Misplaced candidate is 0.333, found at 0',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_index_negative_zero_ok():
    text = """\
    -0;D12L;;;0.8;1;;;NULL;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        ' OK index (0) requested, accepted as (0)',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_index_negative_wun_ok():
    text = """\
    -1;D12L;;;0.8;1;;;NULL;s
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'D12L',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': 's',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        ' OK index (-1) requested, accepted as (-1)',  # Maybe unwanted behavior TODO(sthagen)
        'Conflicting index rules. Failing candidate is (-1), reason is LT_ZERO',
        'Index positions not ordered. Misplaced candidate is -1, found at 0',
    ]

    assert dsl.parse(text) == (ast, diag)


def test_parse_too_many_axes_ok():
    text = """\
    ;A1;;;0.8;1;;;NULL;
    ;A2;;;0.8;1;;;NULL;
    ;A3;;;0.8;1;;;NULL;
    ;A4;;;0.8;1;;;NULL;
    ;A5;;;0.8;1;;;NULL;
    ;A6;;;0.8;1;;;NULL;
    ;A7;;;0.8;1;;;NULL;
    ;A8;;;0.8;1;;;NULL;
    ;A9;;;0.8;1;;;NULL;
    ;A10;;;0.8;1;;;NULL;
    ;A11;;;0.8;1;;;NULL;
    ;A12;;;0.8;1;;;NULL;
    ;A13;;;0.8;1;;;NULL;
    ;A14;;;0.8;1;;;NULL;
    ;A15;;;0.8;1;;;NULL;
    ;A16;;;0.8;1;;;NULL;
    ;A17_is_too_much;;;0.8;1;;;NULL;
    """

    ast = [
        {
            'AXIS_INDEX': 0,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A1',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 1,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A2',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 2,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A3',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 3,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A4',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 4,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A5',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 5,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A6',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 6,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A7',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 7,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A8',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 8,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A9',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 9,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A10',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 10,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A11',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 11,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A12',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 12,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A13',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 13,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A14',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 14,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A15',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
        {
            'AXIS_INDEX': 15,
            'AXIS_LIMIT': 0.8,
            'AXIS_LIMIT_FOLDED': False,
            'AXIS_MAX': 1,
            'AXIS_META': '',
            'AXIS_MIN': 0.4666666666666668,
            'AXIS_MIN_FOLDED': False,
            'AXIS_NAME': 'A16',
            'AXIS_TYPE': 'LINEAR',
            'AXIS_UNIT': '1',
            'AXIS_VALUE': 'NULL',
        },
    ]

    diag = [
        '17 dimensions requested, but only 16 accepted. Maximum is 16',
    ]

    assert dsl.parse(text) == (ast, diag)


@given(a_mi=st.floats(), a_li=st.floats(), a_ma=st.floats(), a_va=st.floats())
def test_parse_bimonotone_floats_stat(a_mi, a_li, a_ma, a_va):
    text = f"""\
    ;DXL;BIMONOTONE;{a_mi};{a_li};{a_ma};;;{a_va};%;SHOW_MIN
    """

    diag = []

    ([axis], diagnoses) = dsl.parse(text)
    assert axis['AXIS_INDEX'] == 0
    assert isinstance(axis['AXIS_LIMIT'], (float, int))
    assert axis['AXIS_LIMIT_FOLDED'] is False
    assert isinstance(axis['AXIS_MAX'], (float, int))
    assert axis['AXIS_META'] == 'SHOW_MIN'
    assert isinstance(axis['AXIS_MIN'], (float, int))
    assert axis['AXIS_MIN_FOLDED'] is False
    assert axis['AXIS_NAME'] == 'DXL'
    assert axis['AXIS_TYPE'] == 'BIMONOTONE'
    assert axis['AXIS_UNIT'] == '%'
    assert isinstance(axis['AXIS_VALUE'], (float, int))
    assert diagnoses == diag



from typing import no_type_check

from PIL import Image, ImageDraw, ImageFont  # type: ignore

import piemap.geometry as geom
from piemap import (
    ANGLE_MAX,
    ANGLE_OFF,
    CIRCLE_CENTER,
    GRAY,
    GREEN,
    HEIGHT,
    HEIGHT_OFF,
    LABEL_SIZE,
    LINE_WIDTH,
    PIE_BOX,
    RADIUS,
    RED,
    SUBTITLE_SIZE,
    TITLE_SIZE,
    TOP_LEFT_X,
    TOP_LEFT_Y,
    WHITE,
    WIDTH,
    WIDTH_HALF,
    YELLOW,
)

FONT_PATH = '../FreeMono.ttf'
OUT_PNG_PATH = '../bitmap.png'
FAKE_EPS_DEGREES = 0.05

Coord = tuple[float, float]
BBox = tuple[Coord, Coord]


@no_type_check
def start_of(n: int, n_max: int) -> float:
    """Start angle in values [0, MAX] first ref is 12 o'clock (OFF) with the of sector n of [1, N]."""
    fractional_angle = ANGLE_MAX / n_max
    return (n * fractional_angle + ANGLE_OFF - fractional_angle / 2) % ANGLE_MAX


@no_type_check
def middle_of(n: int, n_max: int) -> float:
    """Middle angle (axis) in values [0, MAX] first ref is 12 o'clock (OFF) with the of sector n of [1, N]."""
    return (n * ANGLE_MAX / n_max + ANGLE_OFF) % ANGLE_MAX


@no_type_check
def end_of(n: int, n_max: int) -> float:
    """End angle in values [0, MAX] first ref is 12 o'clock (OFF) with the of sector n of [1, N]."""
    return (start_of(n, n_max) + ANGLE_MAX / n_max) % ANGLE_MAX


@no_type_check
def bbox_from_radii(ref_r: int, r: float) -> BBox:
    """Return bbox as left upper and right lower x,y tuples from radius within reference radius"""
    shift = ref_r - r
    return (TOP_LEFT_X + shift, TOP_LEFT_Y + shift), (TOP_LEFT_X + ref_r + r, TOP_LEFT_Y + ref_r + r)


@no_type_check
def extrude_bbox_by(bbox: BBox, extrusion: int) -> BBox:
    """Extrude bounding box symmetrically by extrusion amount."""
    return (bbox[0][0] - extrusion, bbox[0][1] - extrusion), (bbox[1][0] + extrusion, bbox[1][1] + extrusion)


@no_type_check
def fake_full_circle(angle: float) -> Coord:
    """Fake a full circle to reduce the number of functions used."""
    return angle, angle - FAKE_EPS_DEGREES


@no_type_check
def fake_line(angle: float) -> Coord:
    """Fake a line to reduce the number of functions used."""
    return angle, angle + FAKE_EPS_DEGREES


@no_type_check
def draw_titles(engine, title: str, subtitle: str) -> None:
    """Draw the title and subtitle centered on top of the pie and adjust placement for text length."""
    t_fnt = ImageFont.truetype(FONT_PATH, TITLE_SIZE)
    st_fnt = ImageFont.truetype(FONT_PATH, SUBTITLE_SIZE)
    t_len = engine.textlength(title, font=t_fnt)
    st_len = engine.textlength(subtitle, font=st_fnt)
    engine.multiline_text((WIDTH_HALF - int(t_len / 2), HEIGHT_OFF), title, font=t_fnt, fill=(0, 0, 0), align='center')
    engine.multiline_text(
        (WIDTH_HALF - int(st_len / 2), HEIGHT_OFF + TITLE_SIZE), subtitle, font=st_fnt, fill=(0, 0, 0), align='center'
    )


@no_type_check
def draw_label_at(engine, label: str, at: Coord) -> None:
    """Draw the label starting at position."""
    fnt = ImageFont.truetype(FONT_PATH, LABEL_SIZE)
    l_len = engine.textlength(label, font=fnt)
    engine.multiline_text((at[0] - int(l_len / 2), at[1]), label, font=fnt, fill=(0, 0, 0), align='center')


@no_type_check
def to_plot(xy: Coord) -> Coord:
    """Map [0, 1] x [0, 1] onto the plot area that goes east, south."""
    return NotImplemented


@no_type_check
def render(values: tuple[float | None, ...]) -> None:
    """Make importable to support tests."""
    n_dim = len(values)

    # create an image for prototyping
    im = Image.new(mode='RGBA', size=(WIDTH, HEIGHT), color=WHITE)

    # create prototyping draw context
    draw = ImageDraw.Draw(im)

    # All good if above disc at 80% all sectors below receive proportional red coloring adding to the yellow
    draw.pieslice(bbox_from_radii(RADIUS, RADIUS * 0.80), *fake_full_circle(ANGLE_OFF), fill=RED)

    # Inner disc to hide singularity noise at center
    draw.pieslice(bbox_from_radii(RADIUS, RADIUS * 0.03), *fake_full_circle(ANGLE_OFF), fill=GRAY)

    # The sectors
    for n, val in enumerate(values):
        if val is None:
            draw.pieslice(bbox_from_radii(RADIUS, RADIUS), start_of(n, n_dim), end_of(n, n_dim), fill=WHITE)
            continue

        color = YELLOW if val < RADIUS * 0.80 else GREEN
        draw.pieslice(bbox_from_radii(RADIUS, val), start_of(n, n_dim), end_of(n, n_dim), fill=color)

    draw_label_at(draw, '^', CIRCLE_CENTER)  # protonly
    draw_label_at(draw, 'L', (CIRCLE_CENTER[0] - RADIUS, CIRCLE_CENTER[1] + RADIUS))  # protonly
    # The axes and dim=value labels
    for n, val in enumerate(values):
        draw.pieslice(
            extrude_bbox_by(PIE_BOX, 15), *fake_line(middle_of(n, n_dim)), fill=GRAY, outline=GRAY, width=LINE_WIDTH
        )
        coords = geom.xy_point_from_radius_angle(RADIUS, middle_of(n, n_dim), *CIRCLE_CENTER)
        if val is None:
            val_disp = 'n/a'
        else:
            val_disp = f'{round(val, 1)}|{round(val/RADIUS, 1)}'  # |{[round(c, 1) for c in coords]}'
            # ang_quest = (middle_of(n, n_dim) - ANGLE_OFF) % 360
            # val_disp = f'{round(ang_quest, 1)}|{[round(c, 1) for c in coords]}'
        # draw_label_at(draw, f'd[{n}]={val_disp}', (PIE_BOX[0][0], PIE_BOX[0][1] + n * LABEL_SIZE))
        draw_label_at(draw, f'd[{n}]={val_disp}', coords)

    # outer circle (axis marker joining all dimensions)
    draw.pieslice(
        bbox_from_radii(RADIUS, RADIUS * 1.00), *fake_full_circle(ANGLE_OFF), fill=None, outline=GRAY, width=LINE_WIDTH
    )

    # All good if above disc at 80% circle only as marker
    draw.pieslice(
        bbox_from_radii(RADIUS, RADIUS * 0.80), *fake_full_circle(ANGLE_OFF), fill=None, outline=GRAY, width=1
    )

    # title and sub title
    draw_titles(draw, title='Hällo', subtitle='Wörldß')
    del draw

    # im.resize((640, 640), resample=Image.Resampling.LANCZOS)
    im.save(OUT_PNG_PATH, 'PNG')

    im.show()


if __name__ == '__main__':  # pragma: no cover
    R = RADIUS
    r = (R, R * 0.90, R * 0.80, R * 0.70, R * 0.65, R * 0.60, R * 0.55, 0, None)
    render(r)



"""CLI interface for single view visualization mapping parameter values to an area scale resembling a (quality) pie."""
import argparse
import sys

import piemap.api as api
from piemap import APP_ALIAS, APP_NAME, KNOWN_FORMATS, parse_csl


def parse_request(argv: list[str]) -> int | argparse.Namespace:
    """DRY."""
    parser = argparse.ArgumentParser(
        prog=APP_ALIAS, description=APP_NAME, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        '--input-path',
        '-i',
        dest='input_path',
        default='',
        help='the data (format DSL) to create the quality pie. Optional\n(default: positional input path value)',
        required=False,
    )
    parser.add_argument(
        'input_path_pos',
        nargs='?',
        default='',
        help='path to data file.',
    )
    parser.add_argument(
        '--out-path',
        '-o',
        dest='out_path',
        default='piemap',
        help='output file path for graphic (default: piemap.png)',
    )
    parser.add_argument(
        '--formats',
        '-f',
        dest='format_type_csl',
        default='png',
        help='formats (jpeg, png, svg) as comma separated list for taxonomy (default: png)',
    )
    parser.add_argument(
        '--dry-run',
        '-d',
        dest='dry_run',
        default=False,
        action='store_true',
        help='only log what would be done - no real actions (default: False)',
    )
    parser.add_argument(
        '--quiet',
        '-q',
        dest='quiet',
        default=False,
        action='store_true',
        help='work as quiet as possible (default: False)',
    )
    parser.add_argument(
        '--verbose',
        '-v',
        dest='verbose',
        default=False,
        action='store_true',
        help='work logging more information along the way (default: False)',
    )
    if not argv:
        parser.print_help()
        return 0

    options = parser.parse_args(argv)

    if options.verbose and options.quiet:
        parser.error('you cannot be quiet and verbose at the same time')

    format_types = parse_csl(options.format_type_csl)
    for fmt in format_types:
        if fmt not in KNOWN_FORMATS:
            parser.error(f'requested format {fmt} for quality pie not in {KNOWN_FORMATS}')

    if not options.input_path:
        if options.input_path_pos:
            options.input_path = options.input_path_pos
        else:
            options.input_path = 'piemap.dsl'

    return options


def main(argv: list[str] | None = None) -> int:
    """Delegate processing to functional module."""
    argv = sys.argv[1:] if argv is None else argv
    options = parse_request(argv)
    if isinstance(options, int):
        return 0
    return api.main(options)



import hypothesis.strategies as st
from hypothesis import given

import piemap.bitmap as bitm


def test_start_of():
    assert bitm.start_of(0, 1) == (270 - 360 / 1 / 2) % 360


def test_middle_of():
    assert bitm.middle_of(0, 1) == 270


def test_end_of():
    assert bitm.end_of(0, 1) == (270 + 360 / 1 / 2) % 360


def test_bbox_from_radii_extreme_zero():
    assert bitm.bbox_from_radii(1, 0) == ((91.0, 141.0), (91.0, 141.0))


def test_bbox_from_radii_boring_same():
    assert bitm.bbox_from_radii(1, 1) == ((90.0, 140.0), (92.0, 142.0))


@given(n=st.integers(min_value=0, max_value=100))
def test_start_of_data_driven(n):
    n_max = n + 1
    assert 0 <= bitm.start_of(n, n_max) <= 360


@given(angle=st.integers(min_value=0, max_value=100))
def test_fake_full_circle(angle):
    assert bitm.fake_full_circle(angle) == (angle, angle - bitm.FAKE_EPS_DEGREES)


@given(angle=st.integers(min_value=0, max_value=100))
def test_fake_line(angle):
    assert bitm.fake_line(angle) == (angle, angle + bitm.FAKE_EPS_DEGREES)



import math
from typing import no_type_check


@no_type_check
def exact_top_of_center(angle):
    """
    True, if angle leads to point exactly top of center
    """
    return True if angle == 270 else False


@no_type_check
def exact_bottom_of_center(angle):
    """
    True, if angle leads to point exactly bottom of center
    """
    return True if angle == 90 else False


@no_type_check
def right_bottom_of_center(angle):
    """
    True, if angle leads to point right bottom of center
    """
    return True if angle < 90 else False


@no_type_check
def right_top_of_center(angle):
    """
    True, if angle leads to point right top of center
    """
    return True if angle > 270 else False


@no_type_check
def left_bottom_of_center(angle):
    """
    True, if angle leads to point left bottom of center
    """
    return True if 90 < angle < 180 else False


@no_type_check
def left_top_of_center(angle):
    """
    True, if angle leads to point left top of center
    """
    return True if 180 <= angle < 270 else False


@no_type_check
def octant_of_angle(angle):
    """
    derive octant in (N,NW,W,SW,S,SE,E,NE) counterclockwise from angle
    """
    try:
        angle = int(angle)
    except TypeError:
        return 'X'
    if exact_top_of_center(angle):
        return 'N'
    elif exact_bottom_of_center(angle):
        return 'S'
    elif right_bottom_of_center(angle):
        return 'E' if angle == 0 else 'SE'
    elif left_top_of_center(angle):
        return 'W' if angle == 180 else 'NW'
    elif left_bottom_of_center(angle):
        return 'SW'
    elif right_top_of_center(angle):
        return 'NE'


@no_type_check
def axis_name_circle_adjust(angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep=None):
    """
    calculate axis name placements from bounding box and angle
    """

    def bounding_box_from_font(fnt_size_pts, txt_angle, fnt_name, axs_name):
        """
        TODO(sthagen): PHP::imagettfbbox(font_size_pts, text_angle, font_name, axis_name)
        6,7	upper left corners X,Y-Pos and 4,5	upper right corners X,Y-Pos
        0,1	lower left corners X,Y-Pos and 2,3	lower right corners X,Y-Pos

        Returns tuple with:
             0,         1,         2,          3,          4,          5,          6,         7
             lo_left_x, lo_left_y, lo_right_x, lo_right_y, hi_right_x, hi_right_y, hi_left_x, hi_left_y
        """
        return 100, 200, 300, None, None, None, None, 400  # TODO(sthagen) implement real functionality

    axis_name_space_sep = True if axis_name_space_sep is None else axis_name_space_sep

    bbox = bounding_box_from_font(font_size_pts, text_angle, font_name, axis_name)
    text_width = abs(bbox[2] - bbox[0])
    text_height = abs(bbox[7] - bbox[1])

    sep_x = 0
    sep_y = 0
    if axis_name_space_sep:
        sep_x = 5
        sep_y = 5

    dx_pixel = 0
    dy_pixel = 0
    octant = octant_of_angle(angle)
    if octant == 'SE':
        dx_pixel = 0 + sep_x
        dy_pixel = +text_height + sep_y
    elif octant == 'NE':
        dx_pixel = 0 + sep_x
        dy_pixel = 0 - sep_y
    elif octant == 'SW':
        dx_pixel = -text_width - sep_x
        dy_pixel = +text_height + sep_y
    elif octant == 'NW':
        dx_pixel = -text_width - sep_x
        dy_pixel = 0 - sep_y
    elif octant == 'S':
        dx_pixel = -text_width / 2
        dy_pixel = +text_height + 2 * sep_y
    elif octant == 'N':
        # exactly on top of center
        dx_pixel = -text_width / 2
        dy_pixel = 0 - 2 * sep_y
    elif octant == 'W':
        dx_pixel = -text_width - sep_x
        dy_pixel = +text_height / 2 + 0
    elif octant == 'E':
        dx_pixel = +text_width + sep_x
        dy_pixel = +text_height / 2 + 0

    # Note: Unmatched octant returns (0,0)
    return dx_pixel, dy_pixel


@no_type_check
def xy_point_from_radius_angle(radius, angle, c_x=0, c_y=0):
    """
    return point as tuple in (x,y) = (radius,angle/[deg]) possibly shifted (c_x, c_y)
    """

    def deg2rad(deg):
        """TODO(sthagen) refactor later."""
        return deg * math.pi / 180.0

    if math.isfinite(angle):
        angle_in_rad = deg2rad(math.fmod(angle + 360, 360))
        x = c_x + math.cos(angle_in_rad) * radius
        y = c_y + math.sin(angle_in_rad) * radius
        return x, y
    else:
        return math.nan, math.nan


@no_type_check
def segment_angle_map(needed_number_of_axis):
    """
    return map of clockwise allocated angles for the needed axis segments
    from imagefilledarc: 0 degrees is located at the three-oclock position,
    ... and the arc is drawn clockwise.
    but we want one $angleMid allways point to 12 oclock, the rest
    ... determined by $neededNumberOfAxis
    and remember, we want the axis to point at 12 oclock, not start-stop
    so n = 1 gives (0,360,360) in theory 180 full circle to 180
       n = 2 gives [(270,90,360), (90,270,180)]
       n = 3 gives [(300,60,360), (60,180,120), (180,300,240)]
    """
    closure_guard, norm = 0, 360

    if needed_number_of_axis == 1:
        return [(closure_guard, norm, norm)]  # early exit for cornercase

    angle_per_sect = norm / needed_number_of_axis
    signed_correct_axis_shift = -angle_per_sect / 2
    triplets = []
    for i in range(0, needed_number_of_axis):
        angle_start = math.fmod(i * angle_per_sect + signed_correct_axis_shift + norm, norm)
        delta = angle_start + angle_per_sect

        angle_stop = derive_stop(angle_start, closure_guard, delta, norm)
        angle_mid = derive_mid(angle_start, angle_stop, norm)

        triplets.append((angle_start, angle_stop, angle_mid))

    return triplets


@no_type_check
def transform_angle_map_ncw_icw(segment_angle_map_ncw):
    """
    return map of adjusted angles for imagefilledarcClockWise from nativeCW
    from imagefilledarc: 0 degrees is located at the three-oclock position,
    ... and the arc is drawn clockwise.
    but we want one $angleMid allways point to 12 oclock, the rest
    ... determined by $neededNumberOfAxis
    so transform maps 360 deg and 0 deg to 270 deg and 90 deg to 0 deg,
    ... i.e. d=-90 modulo 360
    """
    my_eps = 0.0  # 0.0000000000001 -  one zero more and 12+03 oclock for n=1

    if len(segment_angle_map_ncw) == 1:
        return [(270 + my_eps, 270 - my_eps, 270)]  # early exit for cornercase

    closure_guard, norm = 0, 360
    signed_shift_degrees = -90
    return rotate(closure_guard, norm, segment_angle_map_ncw, signed_shift_degrees)


@no_type_check
def transform_angle_map_icw_ncw(segment_angle_map_icw):
    """
    return map of canonicalized angles for nativeClockWise from imagefilledarcCW
    from imagefilledarc: 0 degrees is located at the three-oclock position,
    ... and the arc is drawn clockwise.
    but we want one $angleMid allways point to 12 oclock, the rest
    ... determined by $neededNumberOfAxis
    so transform maps 360 deg and 0 deg to 270 deg and 90 deg to 0 deg,
    ... i.e. d=-90 modulo 360
    """
    closure_guard, norm = 0, 360
    # TODO(sthagen) remove after test: my_eps = 0.0000001

    if len(segment_angle_map_icw) == 1:
        return [(closure_guard, norm, norm)]  # early exit for cornercase

    signed_shift_degrees = +90
    return rotate(closure_guard, norm, segment_angle_map_icw, signed_shift_degrees)


@no_type_check
def rotate(closure_guard, norm, segment_angle_map_icw, signed_shift_degrees):
    """Derive triplet mappings for segments."""
    triplets = []
    for data in segment_angle_map_icw:
        angle_start, angle_stop, angle_mid = data

        angle_start = math.fmod(angle_start + signed_shift_degrees + norm, norm)
        delta = angle_stop + signed_shift_degrees

        angle_stop = derive_stop(angle_start, closure_guard, delta, norm)
        angle_mid = derive_mid(angle_start, angle_stop, norm)

        triplets.append((angle_start, angle_stop, angle_mid))

    return triplets


@no_type_check
def derive_stop(angle_start, closure_guard, delta, norm):
    """Derive stop angle respecting closure via norm."""
    angle = math.fmod(delta + norm, norm)
    if angle < angle_start and angle == closure_guard:
        angle = norm
    return angle


@no_type_check
def derive_mid(angle_start, angle_stop, norm):
    """Derive mid angle respectng closure via norm."""
    angle = (angle_stop + angle_start) / 2.0
    if angle_stop < angle_start:
        angle = (angle_stop + norm + angle_start) / 2.0
    return angle



"""One-view visualization of grouped characterizations (Quality Pie)."""
import datetime as dti
import logging
import os
import pathlib
from typing import no_type_check

# [[[fill git_describe()]]]
__version__ = '2023.10.22+parent.ed38dfeb'
# [[[end]]] (checksum: db573b62f771dae6170699431794e7d5)
__version_info__ = tuple(
    e if '-' not in e else e.split('-')[0] for part in __version__.split('+') for e in part.split('.') if e != 'parent'
)
__all__: list[str] = [
    'ANGLE_MAX',
    'ANGLE_OFF',
    'CIRCLE_CENTER',
    'DARKRED',
    'DEBUG',
    'ENCODING',
    'ENCODING_ERRORS_POLICY',
    'GRAY',
    'GREEN',
    'HEIGHT',
    'HEIGHT_HALF',
    'HEIGHT_OFF',
    'LABEL_SIZE',
    'LINE_WIDTH',
    'PIE_BOX',
    'RADIUS',
    'RADIUS_HALF',
    'RED',
    'SUBTITLE_SIZE',
    'TITLE_SIZE',
    'TOP_LEFT_X',
    'TOP_LEFT_Y',
    'WHITE',
    'WIDTH',
    'WIDTH_HALF',
    'WIDTH_OFF',
    'YELLOW',
    'YELLOWGREEN',
    'log',
]

APP_ALIAS = str(pathlib.Path(__file__).parent.name)
APP_ENV = APP_ALIAS.upper()
APP_NAME = locals()['__doc__']
DEBUG = bool(os.getenv(f'{APP_ENV}_DEBUG', ''))
VERBOSE = bool(os.getenv(f'{APP_ENV}_VERBOSE', ''))
QUIET = False
STRICT = bool(os.getenv(f'{APP_ENV}_STRICT', ''))
ENCODING = 'utf-8'
ENCODING_ERRORS_POLICY = 'ignore'
DEFAULT_CONFIG_NAME = f'.{APP_ALIAS}.json'

COLON = ':'
COMMA = ','
DEFAULT_LF_ONLY = 'YES'
DOT = '.'
KNOWN_FORMATS = ('jpeg', 'png', 'svg')
PIPE = '|'
SEMI = ';'

log = logging.getLogger()  # Module level logger is sufficient
LOG_FOLDER = pathlib.Path('logs')
LOG_FILE = f'{APP_ALIAS}.log'
LOG_PATH = pathlib.Path(LOG_FOLDER, LOG_FILE) if LOG_FOLDER.is_dir() else pathlib.Path(LOG_FILE)
LOG_LEVEL = logging.INFO

TS_FORMAT_LOG = '%Y-%m-%dT%H:%M:%S'
TS_FORMAT_PAYLOADS = '%Y-%m-%d %H:%M:%S.%f UTC'

WIDTH = 1080
HEIGHT = WIDTH
DIAMETER = 900
WIDTH_HALF = WIDTH / 2
HEIGHT_HALF = HEIGHT / 2
RADIUS = DIAMETER / 2
WIDTH_OFF = 0
HEIGHT_OFF = 50
TOP_LEFT_X = WIDTH_HALF - RADIUS
TOP_LEFT_Y = HEIGHT_HALF - RADIUS + HEIGHT_OFF
PIE_BOX = ((TOP_LEFT_X, TOP_LEFT_Y), (TOP_LEFT_X + DIAMETER, TOP_LEFT_Y + DIAMETER))
CIRCLE_CENTER = (WIDTH_HALF, HEIGHT_HALF + HEIGHT_OFF)

ANGLE_OFF = 270
ANGLE_MAX = 360

# font sizes and line width
TITLE_SIZE = 40
SUBTITLE_SIZE = 24
LABEL_SIZE = 16
LINE_WIDTH = 1

# colors
DARKRED = '#900000'
GRAY = '#c0c0c0'
GREEN = '#008800'
RED = '#dd0000'
WHITE = '#ffffff'
YELLOW = '#ffff20'
YELLOWGREEN = '#77ff20'


def parse_csl(csl: str) -> list[str]:
    """DRY."""
    return [fmt.strip().lower() for fmt in csl.split(COMMA) if fmt.strip()]


@no_type_check
def formatTime_RFC3339(self, record, datefmt=None):  # noqa
    """HACK A DID ACK we could inject .astimezone() to localize ..."""
    return dti.datetime.fromtimestamp(record.created, dti.timezone.utc).isoformat()  # pragma: no cover


@no_type_check
def init_logger(name=None, level=None):
    """Initialize module level logger"""
    global log  # pylint: disable=global-statement

    log_format = {
        'format': '%(asctime)s %(levelname)s [%(name)s]: %(message)s',
        'datefmt': TS_FORMAT_LOG,
        # 'filename': LOG_PATH,
        'level': LOG_LEVEL if level is None else level,
    }
    logging.Formatter.formatTime = formatTime_RFC3339
    logging.basicConfig(**log_format)
    log = logging.getLogger(APP_ENV if name is None else name)
    log.propagate = True


init_logger(name=APP_ENV, level=logging.DEBUG if DEBUG else None)



from typing import no_type_check


@no_type_check
def min_from_limit_max(limit, maximum):
    """
    Testdata: (-6, -1, 2, 5, 10) or (0, 5, 8, 11, 16)
    yield (8*(-1) - 5*2) / 3 = -6 or (8*5 - 5*8) / 3 = 0
    a ---- b -- c and bc/ac = 3/8 => 8c - 8b = 3c - 3a => a = (8b - 5c)/3
    """
    return (8.0 * limit - 5.0 * maximum) / 3.0


@no_type_check
def domain_from_limit_max(limit, maximum):
    """Prepare the scales paradigm for non-folded axes."""
    return min_from_limit_max(limit, maximum), maximum


@no_type_check
def limit_folded_from_limit_max(limit, maximum):
    """
    Testdata: (-6, -1, 2, 5, 10) or (0, 5, 8, 11, 16)
    yield 2 * 2 - (-1) = 5 or 2 * 8 - 5 = 11
    """
    return 2.0 * maximum - limit  # explicit maximum + ( maximum - limit )


@no_type_check
def min_folded_from_limit_max(limit, maximum):
    """
    Testdata: (-6, -1, 2, 5, 10) or (0, 5, 8, 11, 16)
    yield (11*2 - 8*(-1)) / 3 = 10 or (11*8 - 8*5) / 3 = 16
    a ---- b -- c ------- e and bc/ce = 3/8 => 8c - 8b = 3e - 3c => a = (11c - 8b)/3
    """
    return (11.0 * maximum - 8.0 * limit) / 3.0


@no_type_check
def domain_folded_from_limit_max(limit, maximum):
    """Prepare the scales paradigm for folded axes."""
    max_domain = min_folded_from_limit_max(limit, maximum)
    min_domain = maximum + maximum - max_domain  # maximum - (max_domain - maximum)
    return min_domain, max_domain


@no_type_check
def value_folded_from_limit_max(value, maximum):
    """
    Testdata: (-6, -1, 2, 5, 10) or (0, 5, 8, 11, 16)  # FIXME COPY
    yield 2 * 2 - (-1) = 5 or 2 * 8 - 5 = 11  # FIXME COPY
    """
    return 2.0 * maximum - value  # explicit maximum - ( value - maximum )


@no_type_check
def limit_ordered_from_domain(domain):
    """
    Testdata: ("3", "2c", "2b", "2a", "1") or ("ORIGIN", "not ok", "LIMIT_VALUE", "ok")
    yield domain[len(domain) // 2] -> "2b" or "LIMIT_VALUE" in domain -> "LIMIT_VALUE"
    """
    if not domain:
        return 'NULL'
    if 'LIMIT_VALUE' in domain:
        return 'LIMIT_VALUE'
    return domain[len(domain) // 2]


@no_type_check
def min_ordered_from_domain(domain):
    """
    Testdata: ("3", "2c", "2b", "2a", "1") or ("ORIGIN", "not ok", "LIMIT_VALUE", "ok")
    yield domain[0] -> "3" or domain[0] -> "ORIGIN"
    """
    return domain[0] if domain else 'NULL'


@no_type_check
def domain_ordered_from_domain(domain):
    """Prepare the scales paradigm for ordered axes."""
    return domain if domain else 'NULL'


@no_type_check
def max_ordered_from_domain(domain):
    """
    Testdata: ("3", "2c", "2b", "2a", "1") or ("ORIGIN", "not ok", "LIMIT_VALUE", "ok")
    yield domain[-1] -> "1" or domain[-1] -> "ok"
    """
    return domain[-1] if domain else 'NULL'



import pytest  # type: ignore

import piemap.cli as cli


def test_main_ok_funny_arg():
    job = ['[]']
    assert cli.main(job) == 0


def test_main_ok_empty_array():
    job = []
    assert cli.main(job) == 0


def test_main_nok_wrong_type_int():
    bad = 42
    message = r"'int' object is not iterable"
    with pytest.raises(TypeError, match=message):
        cli.main(bad)  # type: ignore


def test_parse_request_empty():
    assert cli.parse_request([]) == 0


def test_parse_request_contradict(capsys):
    message = '2'
    with pytest.raises(SystemExit, match=message):
        cli.parse_request(['-q', '-v'])  # type: ignore
    out, err = capsys.readouterr()
    assert not out
    assert 'you cannot be quiet and verbose at the same time' in err


def test_parse_request_invalid_format(capsys):
    message = '2'
    alien = 'alien'
    with pytest.raises(SystemExit, match=message):
        cli.parse_request(['-f', f'jpeg,png,{alien},svg'])  # type: ignore
    out, err = capsys.readouterr()
    assert not out
    assert f'requested format {alien} for quality pie not in {cli.KNOWN_FORMATS}' in err


def test_parse_request_default_out(capsys):
    options = cli.parse_request(['-i', 'piemap.dsl'])  # type: ignore
    assert options.input_path == 'piemap.dsl'
    assert not options.verbose
    assert not options.quiet
    assert options.format_type_csl == 'png'
    assert options.out_path == 'piemap'


def test_parse_request_pos_in(capsys):
    options = cli.parse_request(['piemap.dsl'])  # type: ignore
    assert options.input_path == 'piemap.dsl'


def test_parse_request_default_dsl(capsys):
    options = cli.parse_request(['-o', 'bizarre'])  # type: ignore
    assert options.input_path == 'piemap.dsl'
    assert options.out_path == 'bizarre'



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
import hypothesis.strategies as st
import pytest  # type: ignore
from hypothesis import given

import piemap.geometry as geom


def test_exact_top_of_center_true_ok():
    assert geom.exact_top_of_center(270) is True


def test_exact_top_of_center_false_ok():
    assert geom.exact_top_of_center(270 + 1) is False


def test_exact_bottom_of_center_true_ok():
    assert geom.exact_bottom_of_center(90) is True


def test_exact_bottom_of_center_false_ok():
    assert geom.exact_bottom_of_center(90 + 1) is False


def test_right_bottom_of_center_true_ok():
    assert geom.right_bottom_of_center(90 - 1) is True


def test_right_bottom_of_center_false_ok():
    assert geom.right_bottom_of_center(90) is False


def test_right_top_of_center_true_ok():
    assert geom.right_top_of_center(270 + 1) is True


def test_right_top_of_center_false_ok():
    assert geom.right_top_of_center(270) is False


def test_left_bottom_of_center_true_ok():
    assert geom.left_bottom_of_center(90 + 1) is True


def test_left_bottom_of_center_false_ok():
    assert geom.left_bottom_of_center(90) is False


def test_left_bottom_of_center_true_other_ok():
    assert geom.left_bottom_of_center(180 - 1) is True


def test_left_bottom_of_center_false_other_ok():
    assert geom.left_bottom_of_center(180) is False


def test_left_top_of_center_true_ok():
    assert geom.left_top_of_center(180) is True


def test_left_top_of_center_false_ok():
    assert geom.left_top_of_center(180 - 1) is False


def test_left_top_of_center_true_other_ok():
    assert geom.left_top_of_center(270 - 1) is True


def test_left_top_of_center_false_other_ok():
    assert geom.left_top_of_center(270) is False


def test_left_top_of_center_bad_argument_nok():
    message = r"'<=' not supported between instances of 'int' and 'object'"
    with pytest.raises(TypeError, match=message):
        geom.left_top_of_center(object())


def test_octant_of_angle_for_n_ok():
    assert geom.octant_of_angle(270) == 'N'


def test_octant_of_angle_for_s_ok():
    assert geom.octant_of_angle(90) == 'S'


def test_octant_of_angle_for_e_ok():
    assert geom.octant_of_angle(0) == 'E'


def test_octant_of_angle_for_se_ok():
    assert geom.octant_of_angle(90 - 1) == 'SE'


def test_octant_of_angle_for_w_ok():
    assert geom.octant_of_angle(180) == 'W'


def test_octant_of_angle_for_nw_ok():
    assert geom.octant_of_angle(270 - 1) == 'NW'


def test_octant_of_angle_for_sw_ok():
    assert geom.octant_of_angle(90 + 1) == 'SW'


def test_octant_of_angle_for_ne_ok():
    assert geom.octant_of_angle(270 + 1) == 'NE'


def test_octant_of_angle_for_x_ok():
    assert geom.octant_of_angle(object()) == 'X'


def test_xy_point_from_radius_angle_minimal_ok():
    r, a, dx, dy = 1, 0, 0, 0
    x, y = 1, 0
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_radius_angle_minimal_shift_x_ok():
    r, a, dx, dy = 1, 0, 1, 0
    x, y = 2, 0
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_radius_angle_minimal_shift_y_ok():
    r, a, dx, dy = 1, 0, 0, 1
    x, y = 1, 1
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_radius_angle_minimal_shift_x_y_ok():
    r, a, dx, dy = 1, 0, 1, 1
    x, y = 2, 1
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_all_zero_ok():
    r, a, dx, dy = 0, 0, 0, 0
    x, y = 0, 0
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_radius_zero_ok():
    r, a, dx, dy = 0, 1, 2, 3
    x, y = 2, 3
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_angle_zero_ok():
    r, a, dx, dy = 1, 0, 2, 3
    x, y = 3, 3
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_angle_negative_and_zero_radius_ok():
    r, a, dx, dy = 0, -1, 2, 3
    x, y = 2, 3
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_xy_point_from_angle_negative_ok():
    r, a, dx, dy = 1000, -360, 25, 14
    x, y = 1025, pytest.approx(14, abs=1e-11)
    assert geom.xy_point_from_radius_angle(r, a, dx, dy) == (x, y)


def test_segment_angle_map_one_ok():
    assert geom.segment_angle_map(1) == [(0, 360, 360)]


def test_segment_angle_map_two_ok():
    assert geom.segment_angle_map(2) == [(270, 90, 360), (90, 270, 180)]


def test_segment_angle_map_three_ok():
    assert geom.segment_angle_map(3) == [(300, 60, 360), (60, 180, 120), (180, 300, 240)]


def test_segment_angle_map_four_ok():
    assert geom.segment_angle_map(4) == [(315, 45, 360), (45, 135, 90), (135, 225, 180), (225, 315, 270)]


def test_segment_angle_map_five_ok():
    assert geom.segment_angle_map(5) == [
        (324, 36, 360),
        (36, 108, 72),
        (108, 180, 144),
        (180, 252, 216),
        (252, 324, 288),
    ]


def test_segment_angle_map_six_ok():
    assert geom.segment_angle_map(6) == [
        (330, 30, 360),
        (30, 90, 60),
        (90, 150, 120),
        (150, 210, 180),
        (210, 270, 240),
        (270, 330, 300),
    ]


def test_segment_angle_map_seven_ok():
    assert geom.segment_angle_map(7) == [
        (pytest.approx(334.2857), pytest.approx(25.7143), 360),
        (pytest.approx(25.7143), pytest.approx(77.1429), pytest.approx(51.4286)),
        (pytest.approx(77.1429), pytest.approx(128.5714), pytest.approx(102.8571)),
        (pytest.approx(128.5714), 180, pytest.approx(154.2857)),
        (180, pytest.approx(231.4286), pytest.approx(205.7143)),
        (pytest.approx(231.4286), pytest.approx(282.8571), pytest.approx(257.1429)),
        (pytest.approx(282.8571), pytest.approx(334.2857), pytest.approx(308.5714)),
    ]


def test_segment_angle_map_eight_ok():
    assert geom.segment_angle_map(8) == [
        (337.5, 22.5, 360),
        (22.5, 67.5, 45),
        (67.5, 112.5, 90),
        (112.5, 157.5, 135),
        (157.5, 202.5, 180),
        (202.5, 247.5, 225),
        (247.5, 292.5, 270),
        (292.5, 337.5, 315),
    ]


def test_segment_angle_map_nine_ok():
    assert geom.segment_angle_map(9) == [
        (340, 20, 360),
        (20, 60, 40),
        (60, 100, 80),
        (100, 140, 120),
        (140, 180, 160),
        (180, 220, 200),
        (220, 260, 240),
        (260, 300, 280),
        (300, 340, 320),
    ]


def test_segment_angle_map_ten_ok():
    assert geom.segment_angle_map(10) == [
        (342, 18, 360),
        (18, 54, 36),
        (54, 90, 72),
        (90, 126, 108),
        (126, 162, 144),
        (162, 198, 180),
        (198, 234, 216),
        (234, 270, 252),
        (270, 306, 288),
        (306, 342, 324),
    ]


def test_segment_angle_map_eleven_ok():
    assert geom.segment_angle_map(11) == [
        (343.6363636363636, 16.363636363636374, 360),
        (16.363636363636374, 49.09090909090912, 32.72727272727275),
        (49.09090909090912, 81.81818181818187, 65.4545454545455),
        (81.81818181818181, 114.5454545454545, 98.18181818181816),
        (114.54545454545456, 147.27272727272725, 130.9090909090909),
        (147.27272727272725, 180, 163.63636363636363),
        (180, 212.72727272727275, 196.36363636363637),
        (212.72727272727275, 245.4545454545455, 229.09090909090912),
        (245.4545454545455, 278.18181818181824, 261.81818181818187),
        (278.18181818181824, 310.909090909091, 294.5454545454546),
        (310.9090909090909, 343.6363636363636, 327.27272727272725),
    ]


def test_segment_angle_map_twelve_ok():
    assert geom.segment_angle_map(12) == [
        (345, 15, 360),
        (15, 45, 30),
        (45, 75, 60),
        (75, 105, 90),
        (105, 135, 120),
        (135, 165, 150),
        (165, 195, 180),
        (195, 225, 210),
        (225, 255, 240),
        (255, 285, 270),
        (285, 315, 300),
        (315, 345, 330),
    ]


def test_segment_angle_map_thirteen_ok():
    tol = 5e-3
    assert geom.segment_angle_map(13) == [
        (pytest.approx(346.1538, abs=tol), pytest.approx(13.8462, abs=tol), 360),
        (pytest.approx(13.8462, abs=tol), pytest.approx(41.5385, abs=tol), pytest.approx(27.6923, abs=tol)),
        (pytest.approx(41.5385, abs=tol), pytest.approx(69.2308, abs=tol), pytest.approx(55.3846, abs=tol)),
        (pytest.approx(69.2308, abs=tol), pytest.approx(96.9239, abs=tol), pytest.approx(83.0769, abs=tol)),
        (pytest.approx(96.9239, abs=tol), pytest.approx(124.6154, abs=tol), pytest.approx(110.7692, abs=tol)),
        (pytest.approx(124.6154, abs=tol), pytest.approx(152.3077, abs=tol), pytest.approx(138.4615, abs=tol)),
        (pytest.approx(152.3077, abs=tol), 180, pytest.approx(166.1538, abs=tol)),
        (180, pytest.approx(207.6923, abs=tol), pytest.approx(193.8462, abs=tol)),
        (pytest.approx(207.6923, abs=tol), pytest.approx(235.3846, abs=tol), pytest.approx(221.5385, abs=tol)),
        (pytest.approx(235.3846, abs=tol), pytest.approx(263.0769, abs=tol), pytest.approx(249.2308, abs=tol)),
        (pytest.approx(263.0769, abs=tol), pytest.approx(290.7692, abs=tol), pytest.approx(276.9231, abs=tol)),
        (pytest.approx(290.7692, abs=tol), pytest.approx(318.4615, abs=tol), pytest.approx(304.6154, abs=tol)),
        (pytest.approx(318.4615, abs=tol), pytest.approx(346.1538, abs=tol), pytest.approx(332.3077, abs=tol)),
    ]


def test_segment_angle_map_fourteen_ok():
    assert geom.segment_angle_map(14) == [
        (347.14285714285717, 12.85714285714289, 360),
        (12.857142857142833, 38.571428571428555, 25.714285714285694),
        (38.571428571428555, 64.28571428571428, 51.428571428571416),
        (64.28571428571428, 90, 77.14285714285714),
        (90, 115.71428571428572, 102.85714285714286),
        (115.71428571428572, 141.42857142857144, 128.57142857142858),
        (141.42857142857144, 167.1428571428571, 154.28571428571428),
        (167.1428571428571, 192.8571428571429, 180),
        (192.8571428571429, 218.57142857142867, 205.71428571428578),
        (218.57142857142856, 244.28571428571422, 231.4285714285714),
        (244.28571428571433, 270, 257.14285714285717),
        (270, 295.7142857142858, 282.8571428571429),
        (295.7142857142858, 321.42857142857156, 308.57142857142867),
        (321.42857142857144, 347.1428571428571, 334.2857142857143),
    ]


def test_segment_angle_map_fifteen_ok():
    assert geom.segment_angle_map(15) == [
        (348, 12, 360),
        (12, 36, 24),
        (36, 60, 48),
        (60, 84, 72),
        (84, 108, 96),
        (108, 132, 120),
        (132, 156, 144),
        (156, 180, 168),
        (180, 204, 192),
        (204, 228, 216),
        (228, 252, 240),
        (252, 276, 264),
        (276, 300, 288),
        (300, 324, 312),
        (324, 348, 336),
    ]


def test_segment_angle_map_sixteen_ok():
    assert geom.segment_angle_map(16) == [
        (348.75, 11.25, 360),
        (11.25, 33.75, 22.5),
        (33.75, 56.25, 45),
        (56.25, 78.75, 67.5),
        (78.75, 101.25, 90),
        (101.25, 123.75, 112.5),
        (123.75, 146.25, 135),
        (146.25, 168.75, 157.5),
        (168.75, 191.25, 180),
        (191.25, 213.75, 202.5),
        (213.75, 236.25, 225),
        (236.25, 258.75, 247.5),
        (258.75, 281.25, 270),
        (281.25, 303.75, 292.5),
        (303.75, 326.25, 315),
        (326.25, 348.75, 337.5),
    ]


def test_transform_angle_map_ncw_icw_minimal_ok():
    assert geom.transform_angle_map_ncw_icw([(0, 360, 360)]) == [(270, 270, 270)]


def test_transform_angle_map_ncw_icw_minimal_idempotent_ok():
    assert geom.transform_angle_map_ncw_icw([(270, 270, 270)]) == [(270, 270, 270)]


def test_transform_angle_map_ncw_icw_two_ok():
    assert geom.transform_angle_map_ncw_icw([(270, 90, 360), (90, 270, 180)]) == [(180, 360, 270), (0, 180, 90)]


def test_transform_angle_map_ncw_icw_two_rotate_once_ok():
    assert geom.transform_angle_map_ncw_icw([(180, 360, 270), (0, 180, 90)]) == [(90, 270, 180), (270, 90, 360)]


def test_transform_angle_map_ncw_icw_two_rotate_twice_ok():
    assert geom.transform_angle_map_ncw_icw([(90, 270, 180), (270, 90, 360)]) == [(0, 180, 90), (180, 360, 270)]


def test_transform_angle_map_ncw_icw_two_rotate_thrice_ok():
    assert geom.transform_angle_map_ncw_icw([(0, 180, 90), (180, 360, 270)]) == [(270, 90, 360), (90, 270, 180)]


def test_transform_angle_map_ncw_icw_three_ok():
    assert geom.transform_angle_map_ncw_icw([(300, 60, 360), (60, 180, 120), (180, 300, 240)]) == [
        (210, 330, 270),
        (330, 90, 390),
        (90, 210, 150),
    ]


def test_transform_angle_map_ncw_icw_three_rotate_once_ok():
    assert geom.transform_angle_map_ncw_icw([(210, 330, 270), (330, 90, 390), (90, 210, 150)]) == [
        (120, 240, 180),
        (240, 360, 300),
        (0, 120, 60),
    ]


def test_transform_angle_map_ncw_icw_three_rotate_twice_ok():
    assert geom.transform_angle_map_ncw_icw([(120, 240, 180), (240, 360, 300), (0, 120, 60)]) == [
        (30, 150, 90),
        (150, 270, 210),
        (270, 30, 330),
    ]


def test_transform_angle_map_ncw_icw_four_ok():
    ingoing = [(315, 45, 360), (45, 135, 90), (135, 225, 180), (225, 315, 270)]
    outgoing = [(225, 315, 270), (315, 45, 360), (45, 135, 90), (135, 225, 180)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_four_rotate_once_ok():
    ingoing = [(225, 315, 270), (315, 45, 360), (45, 135, 90), (135, 225, 180)]
    outgoing = [(135, 225, 180), (225, 315, 270), (315, 45, 360), (45, 135, 90)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_four_rotate_twice_ok():
    ingoing = [(135, 225, 180), (225, 315, 270), (315, 45, 360), (45, 135, 90)]
    outgoing = [(45, 135, 90), (135, 225, 180), (225, 315, 270), (315, 45, 360)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_four_rotate_thrice_ok():
    ingoing = [(45, 135, 90), (135, 225, 180), (225, 315, 270), (315, 45, 360)]
    outgoing = [(315, 45, 360), (45, 135, 90), (135, 225, 180), (225, 315, 270)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_five_ok():
    ingoing = [(324, 36, 360), (36, 108, 72), (108, 180, 144), (180, 252, 216), (252, 324, 288)]
    outgoing = [(234, 306, 270), (306, 18, 342), (18, 90, 54), (90, 162, 126), (162, 234, 198)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_five_rotate_once_ok():
    ingoing = [(234, 306, 270), (306, 18, 342), (18, 90, 54), (90, 162, 126), (162, 234, 198)]
    outgoing = [(144, 216, 180), (216, 288, 252), (288, 360, 324), (0, 72, 36), (72, 144, 108)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_five_rotate_twice_ok():
    ingoing = [(144, 216, 180), (216, 288, 252), (288, 360, 324), (0, 72, 36), (72, 144, 108)]
    outgoing = [(54, 126, 90), (126, 198, 162), (198, 270, 234), (270, 342, 306), (342, 54, 378)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_five_rotate_thrice_ok():
    ingoing = [(54, 126, 90), (126, 198, 162), (198, 270, 234), (270, 342, 306), (342, 54, 378)]
    outgoing = [(324, 36, 360), (36, 108, 72), (108, 180, 144), (180, 252, 216), (252, 324, 288)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_six_ok():
    ingoing = [(330, 30, 360), (30, 90, 60), (90, 150, 120), (150, 210, 180), (210, 270, 240), (270, 330, 300)]
    outgoing = [(240, 300, 270), (300, 360, 330), (0, 60, 30), (60, 120, 90), (120, 180, 150), (180, 240, 210)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_six_rotate_once_ok():
    ingoing = [(240, 300, 270), (300, 360, 330), (0, 60, 30), (60, 120, 90), (120, 180, 150), (180, 240, 210)]
    outgoing = [(150, 210, 180), (210, 270, 240), (270, 330, 300), (330, 30, 360), (30, 90, 60), (90, 150, 120)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_six_rotate_twice_ok():
    ingoing = [(150, 210, 180), (210, 270, 240), (270, 330, 300), (330, 30, 360), (30, 90, 60), (90, 150, 120)]
    outgoing = [(60, 120, 90), (120, 180, 150), (180, 240, 210), (240, 300, 270), (300, 360, 330), (0, 60, 30)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_ncw_icw_six_rotate_thrice_ok():
    ingoing = [(60, 120, 90), (120, 180, 150), (180, 240, 210), (240, 300, 270), (300, 360, 330), (0, 60, 30)]
    outgoing = [(330, 30, 360), (30, 90, 60), (90, 150, 120), (150, 210, 180), (210, 270, 240), (270, 330, 300)]
    assert geom.transform_angle_map_ncw_icw(ingoing) == outgoing


def test_transform_angle_map_icw_ncw_minimal_ok():
    assert geom.transform_angle_map_icw_ncw([(270, 270, 270)]) == [(0, 360, 360)]


def test_transform_angle_map_icw_ncw_minimal_not_really_idempotent_ok():
    assert geom.transform_angle_map_icw_ncw([(270, 270, 270)]) == [(0, 360, 360)]


def test_transform_angle_map_icw_ncw_two_ok():
    assert geom.transform_angle_map_icw_ncw([(180, 360, 270), (0, 180, 90)]) == [(270, 90, 360), (90, 270, 180)]


def test_axis_name_circle_adjust_n_ok():
    angle, font_size_pts, text_angle = 270, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'N'
    assert geom.axis_name_circle_adjust(*args) == (-100, -10)


def test_axis_name_circle_adjust_s_ok():
    angle, font_size_pts, text_angle = 90, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'S'
    assert geom.axis_name_circle_adjust(*args) == (-100, 210)


def test_axis_name_circle_adjust_se_ok():
    angle, font_size_pts, text_angle = 90 - 1, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'SE'
    assert geom.axis_name_circle_adjust(*args) == (5, 205)


def test_axis_name_circle_adjust_ne_ok():
    angle, font_size_pts, text_angle = 270 + 1, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'NE'
    assert geom.axis_name_circle_adjust(*args) == (5, -5)


def test_axis_name_circle_adjust_e_ok():
    angle, font_size_pts, text_angle = 0, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'E'
    assert geom.axis_name_circle_adjust(*args) == (205, 100)


def test_axis_name_circle_adjust_w_ok():
    angle, font_size_pts, text_angle = 180, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'W'
    assert geom.axis_name_circle_adjust(*args) == (-205, 100)


def test_axis_name_circle_adjust_sw_ok():
    angle, font_size_pts, text_angle = 90 + 1, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'SW'
    assert geom.axis_name_circle_adjust(*args) == (-205, 205)


def test_axis_name_circle_adjust_nw_ok():
    angle, font_size_pts, text_angle = 270 - 1, 1, 0
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    assert geom.octant_of_angle(angle) == 'NW'
    assert geom.axis_name_circle_adjust(*args) == (-205, -5)


@given(angle=st.floats(allow_nan=False, allow_infinity=False))
def test_octant_of_angle_float_stat(angle):
    assert geom.octant_of_angle(angle) in ('N', 'S', 'E', 'W', 'NE', 'NW', 'SE', 'SW')


@given(angle=st.floats(allow_nan=False, allow_infinity=False), font_size_pts=st.floats(), text_angle=st.floats())
def test_axis_name_circle_adjust_floats_stat(angle, font_size_pts, text_angle):
    font_name, axis_name, axis_name_space_sep = 'Invisible', 'Look', None
    args = (angle, font_size_pts, text_angle, font_name, axis_name, axis_name_space_sep)
    dx_pix, dy_pix = geom.axis_name_circle_adjust(*args)
    assert isinstance(dx_pix, (float, int))
    assert isinstance(dy_pix, (float, int))


@given(start=st.floats(), mid=st.floats(), stop=st.floats())
def test_transform_angle_map_icw_ncw_floats_stat(start, mid, stop):
    [(x, y, z)] = geom.transform_angle_map_icw_ncw([(start, mid, stop)])
    assert isinstance(x, (float, int))
    assert isinstance(y, (float, int))
    assert isinstance(z, (float, int))


@given(x=st.floats(), y=st.floats(), z=st.floats())
def test_transform_angle_map_ncw_icw_floats_stat(x, y, z):
    [(start, mid, stop)] = geom.transform_angle_map_ncw_icw([(x, y, z)])
    assert isinstance(start, (float, int))
    assert isinstance(mid, (float, int))
    assert isinstance(stop, (float, int))


@given(r=st.floats(), a=st.floats(), dx=st.floats(), dy=st.floats())
def test_xy_point_from_radius_angle_floats_stat(r, a, dx, dy):
    x, y = geom.xy_point_from_radius_angle(r, a, dx, dy)
    assert isinstance(x, (float, int))
    assert isinstance(y, (float, int))



# -*- coding: utf-8 -*-
# pylint: disable=missing-docstring,unused-import,reimported
import hypothesis.strategies as st
from hypothesis import given

import piemap.projections as pr

ORDERED_SAMPLE_ONE = ('3', '2c', '2b', '2a', '1')
ORDERED_SAMPLE_TWO = ('ORIGIN', 'not ok', 'LIMIT_VALUE', 'ok')


def test_min_from_limit_max_sample_one_ok():
    assert pr.min_from_limit_max(-1, 2) == -6


def test_min_from_limit_max_sample_two_ok():
    assert pr.min_from_limit_max(5, 8) == 0


def test_min_from_limit_max_sample_d1f_ok():
    assert pr.min_from_limit_max(12, 15) == 7


def test_domain_from_limit_max_sample_one_ok():
    assert pr.domain_from_limit_max(-1, 2) == (-6, 2)


def test_limit_folded_from_limit_max_sample_one_ok():
    assert pr.limit_folded_from_limit_max(-1, 2) == 5


def test_limit_folded_from_limit_max_sample_two_ok():
    assert pr.limit_folded_from_limit_max(5, 8) == 11


def test_min_folded_from_limit_max_sample_one_ok():
    assert pr.min_folded_from_limit_max(-1, 2) == 10


def test_min_folded_from_limit_max_sample_two_ok():
    assert pr.min_folded_from_limit_max(5, 8) == 16


def test_min_folded_from_limit_max_sample_d1f_ok():
    assert pr.min_folded_from_limit_max(12, 15) == 23


def test_domain_folded_from_limit_max_sample_one_ok():
    assert pr.domain_folded_from_limit_max(-1, 2) == (-6, 10)


def test_domain_folded_from_limit_max_sample_two_ok():
    assert pr.domain_folded_from_limit_max(5, 8) == (0, 16)


def test_domain_folded_from_limit_max_sample_d1f_ok():
    assert pr.domain_folded_from_limit_max(12, 15) == (7, 23)


def test_value_folded_from_limit_max_sample_one_ok():
    assert pr.value_folded_from_limit_max(-1, 2) == 5


def test_value_folded_from_limit_max_sample_two_ok():
    assert pr.value_folded_from_limit_max(5, 8) == 11


def test_limit_ordered_from_domain_sample_one_ok():
    assert pr.limit_ordered_from_domain(ORDERED_SAMPLE_ONE) == '2b'


def test_limit_ordered_from_domain_sample_two_ok():
    assert pr.limit_ordered_from_domain(ORDERED_SAMPLE_TWO) == 'LIMIT_VALUE'


def test_min_ordered_from_domain_sample_one_ok():
    assert pr.min_ordered_from_domain(ORDERED_SAMPLE_ONE) == ORDERED_SAMPLE_ONE[0]


def test_limit_ordered_from_domain_with_limit_value_ok():
    assert pr.limit_ordered_from_domain(('LIMIT_VALUE', 42, False, {})) == 'LIMIT_VALUE'


def test_limit_ordered_from_domain_empty_ok():
    assert pr.limit_ordered_from_domain([]) == 'NULL'


def test_limit_ordered_from_domain_even_ok():
    assert pr.limit_ordered_from_domain([0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0]) == 1


def test_limit_ordered_from_domain_odd_ok():
    assert pr.limit_ordered_from_domain([0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]) == 1


def test_limit_ordered_from_domain_minimal_even_ok():
    assert pr.limit_ordered_from_domain([0, 0, 1, 0]) == 1


def test_limit_ordered_from_domain_minimal_odd_ok():
    assert pr.limit_ordered_from_domain([0, 1, 0]) == 1


def test_min_ordered_from_domain_sample_two_ok():
    assert pr.min_ordered_from_domain(ORDERED_SAMPLE_TWO) == ORDERED_SAMPLE_TWO[0]


def test_min_ordered_from_domain_empty_ok():
    assert pr.min_ordered_from_domain([]) == 'NULL'


def test_domain_ordered_from_domain_sample_one_ok():
    assert pr.domain_ordered_from_domain(ORDERED_SAMPLE_ONE) == ORDERED_SAMPLE_ONE


def test_domain_ordered_from_domain_sample_two_ok():
    assert pr.domain_ordered_from_domain(ORDERED_SAMPLE_TWO) == ORDERED_SAMPLE_TWO


def test_domain_ordered_from_domain_empty_ok():
    assert pr.domain_ordered_from_domain([]) == 'NULL'


def test_max_ordered_from_domain_sample_one_ok():
    assert pr.max_ordered_from_domain(ORDERED_SAMPLE_ONE) == ORDERED_SAMPLE_ONE[-1]


def test_max_ordered_from_domain_sample_two_ok():
    assert pr.max_ordered_from_domain(ORDERED_SAMPLE_TWO) == ORDERED_SAMPLE_TWO[-1]


def test_max_ordered_from_domain_empty_ok():
    assert pr.max_ordered_from_domain([]) == 'NULL'


@given(limit=st.integers(), maximum=st.integers())
def test_min_from_limit_max_int_int(limit, maximum):
    minimum = pr.min_from_limit_max(limit, maximum)
    assert isinstance(minimum, (float, int))


@given(limit=st.floats(), maximum=st.integers())
def test_min_from_limit_max_float_int(limit, maximum):
    minimum = pr.min_from_limit_max(limit, maximum)
    assert isinstance(minimum, (float, int))


@given(limit=st.integers(), maximum=st.floats())
def test_min_from_limit_max_int_float(limit, maximum):
    minimum = pr.min_from_limit_max(limit, maximum)
    assert isinstance(minimum, (float, int))


@given(limit=st.floats(), maximum=st.floats())
def test_min_from_limit_max_float_float(limit, maximum):
    minimum = pr.min_from_limit_max(limit, maximum)
    assert isinstance(minimum, (float, int))


@given(limit=st.integers(), maximum=st.integers())
def test_limit_folded_from_limit_max_int_int(limit, maximum):
    limit_folded = pr.limit_folded_from_limit_max(limit, maximum)
    assert isinstance(limit_folded, (float, int))


@given(limit=st.floats(), maximum=st.integers())
def test_limit_folded_from_limit_max_float_int(limit, maximum):
    limit_folded = pr.limit_folded_from_limit_max(limit, maximum)
    assert isinstance(limit_folded, (float, int))


@given(limit=st.integers(), maximum=st.floats())
def test_limit_folded_from_limit_max_int_float(limit, maximum):
    limit_folded = pr.limit_folded_from_limit_max(limit, maximum)
    assert isinstance(limit_folded, (float, int))


@given(limit=st.floats(), maximum=st.floats())
def test_limit_folded_from_limit_max_float_float(limit, maximum):
    limit_folded = pr.limit_folded_from_limit_max(limit, maximum)
    assert isinstance(limit_folded, (float, int))


@given(limit=st.integers(), maximum=st.integers())
def test_min_folded_from_limit_max_int_int(limit, maximum):
    minimum_folded = pr.min_folded_from_limit_max(limit, maximum)
    assert isinstance(minimum_folded, (float, int))


@given(limit=st.floats(), maximum=st.integers())
def test_min_folded_from_limit_max_float_int(limit, maximum):
    minimum_folded = pr.min_folded_from_limit_max(limit, maximum)
    assert isinstance(minimum_folded, (float, int))


@given(limit=st.integers(), maximum=st.floats())
def test_min_folded_from_limit_max_int_float(limit, maximum):
    minimum_folded = pr.min_folded_from_limit_max(limit, maximum)
    assert isinstance(minimum_folded, (float, int))


@given(limit=st.floats(), maximum=st.floats())
def test_min_folded_from_limit_max_float_float(limit, maximum):
    minimum_folded = pr.min_folded_from_limit_max(limit, maximum)
    assert isinstance(minimum_folded, (float, int))


@given(limit=st.integers(), maximum=st.integers())
def test_value_folded_from_limit_max_int_int(limit, maximum):
    value = pr.value_folded_from_limit_max(limit, maximum)
    assert isinstance(value, (float, int))


@given(limit=st.floats(), maximum=st.integers())
def test_value_folded_from_limit_max_float_int(limit, maximum):
    value = pr.value_folded_from_limit_max(limit, maximum)
    assert isinstance(value, (float, int))


@given(limit=st.integers(), maximum=st.floats())
def test_value_folded_from_limit_max_int_float(limit, maximum):
    value = pr.value_folded_from_limit_max(limit, maximum)
    assert isinstance(value, (float, int))


@given(limit=st.floats(), maximum=st.floats())
def test_value_folded_from_limit_max_float_float(limit, maximum):
    value = pr.value_folded_from_limit_max(limit, maximum)
    assert isinstance(value, (float, int))



"""Create single view visualization mapping parameter values to an area scale resembling a (quality) pie."""
import argparse
from typing import no_type_check

from piemap import DEBUG, ENCODING, ENCODING_ERRORS_POLICY, log


@no_type_check
def parse(text):  # noqa
    """Later we parse the DSL."""
    return NotImplemented


def main(options: argparse.Namespace) -> int:
    log.info(f'{DEBUG=}, {ENCODING=}, {ENCODING_ERRORS_POLICY=}')
    return 0
